@@ -1,60 +1,222 @@
 import argparse
+import logging
 import time
 import requests
 import sys
+import sqlite3
 from bs4 import BeautifulSoup
 
-
 """
-Cambridge is a terminal version of Cambridge Dictionary. Its dictionary data comes from https://dictionary.cambridge.org.
-
-Parsing takes less than 0.4 seconds.
-Most of the time is spent on fetching webpage.
-Total time is mostly between 0.5 to 3 seconds, occasionally 5 secs at most depending on network performance.
+Cambridge is a terminal version of Cambridge Dictionary. 
+The dictionary data comes from https://dictionary.cambridge.org.
 """
 
-DICT_BASE_URL = "https://dictionary.cambridge.org/dictionary/english/"
+CAMBRIDGE_URL = "https://dictionary.cambridge.org"
+DICT_BASE_URL = "https://dictionary.cambridge.org/dictionary/english/"  # if no result found in cambridge, response.url is this.
 SPELLCHECK_BASE_URL = "https://dictionary.cambridge.org/spellcheck/english/?q="
 
+RESQUEST_URL = ""
+RESPONSE_WORD = ""
+RESPONSE_URL = ""
+RESPONSE_TEXT = ""
+
+OP = ["FETCHING", "PARSING"]
+
+def list_words(args, con, cur):
+    # The subparser i.e. the sub-command isn't in the namespace of args
+    if args.delete:
+        word = " ".join(args.delete)
+        if delete_word(con, cur, word):
+            print("'" + word + "'" + " got deleted from cache successfully")
+        else:
+            logger.error("No such word '%s' in cache", word)
+    elif args.random:
+        try:
+            # data is something like [('hello',), ('good',), ('world',)]
+            data = get_random_words(cur)
+        except sqlite3.OperationalError:
+            logger.error("You may haven't searched any word yet")
+        else:
+            for i in data:
+                print(i[0])
+    else:
+        try:
+            # data is something like [('hello',), ('good',), ('world',)]
+            data = get_response_words(cur)
+        except sqlite3.OperationalError:
+            logger.error("You may haven't searched any word yet")
+        else:
+            if args.time:
+                data.sort(reverse=True, key=lambda tup: tup[1])
+            else:
+                data.sort()
+            for i in data:
+                print(i[0])
+
+
+def search_words(args, con, cur):
+    global REQUEST_URL
+    global RESPONSE_WORD
+    global RESPONSE_URL
+    global RESPONSE_TEXT
+
+    input_word, query_word = get_query_word(args)
+    REQUEST_URL = DICT_BASE_URL + query_word
+
+    if args.verbose:
+        logger.setLevel(logging.DEBUG)
+
+    if len(check_table(cur)) == 0:
+        logger.debug("Searching %s for '%s'", CAMBRIDGE_URL, input_word)
+        if request_and_print(REQUEST_URL, args):
+            logger.debug("Caching search result of '%s'", input_word)
+            create_table(con, cur)
+            insert_into_table(con, cur, input_word, RESPONSE_WORD, RESPONSE_URL,RESPONSE_TEXT)
+            logger.debug("Done caching search result of '%s'", input_word)
+    else:
+        data = get_cache(cur, input_word, RESQUEST_URL)
+        if  data is None:
+            logger.debug("Searching %s for '%s'", CAMBRIDGE_URL, input_word)
+            if request_and_print(REQUEST_URL, args):
+                try:
+                    logger.debug("Caching search result of '%s'", input_word)
+                    insert_into_table(con, cur, input_word, RESPONSE_WORD, RESPONSE_URL,RESPONSE_TEXT)
+                # If a user inputs a word with plural form, it won't get cached again
+                except sqlite3.IntegrityError:
+                    logger.debug("Not to cache word of '%s' because it already exists", RESPONSE_WORD)
+                else:
+                    logger.debug("Done caching search result of '%s'", input_word)
+        else:
+            logger.debug("Already cached '%s' before. Use cache", input_word)
+            RESPONSE_URL, RESPONSE_TEXT = data[0], data[1]
+            print_dict()
+
 
 def parse_args():
     parser = argparse.ArgumentParser(
         description="Terminal Version of Cambridge Dictionary"
     )
-    parser.add_argument(
-        "words",
-        nargs="+",  # it makes -w input value in dict data structure even with only one word
-        help="A word or a phrase you want to look up",
+    
+    # Add sub-command capability that can identify different sub-command name 
+    sub_parsers = parser.add_subparsers(dest='subparser_name')
+
+    # Add sub-command l
+    parser_lw = sub_parsers.add_parser("l", help="list all the words you've successfully searched in alphabetical order")
+
+    # Make sub-command l run default funtion of "list_words"
+    parser_lw.set_defaults(func=list_words)
+
+    # Add optional argument for deleting a word from word list
+    parser_lw.add_argument(
+        "-d",
+        "--delete",
+        nargs="+",
+        help="delete a word or phrase from cache",
     )
-    args = parser.parse_args()
+
+    # Add optional argument for listing all words by time
+    parser_lw.add_argument(
+        "-t",
+        "--time",
+        action="store_true",
+        help="list all the words you've successfully searched in reverse chronological order",
+    )
+
+    # Add optional argument for listing words randomly chosen 
+    parser_lw.add_argument(
+        "-r",
+        "--random",
+        action="store_true",
+        help="randomly list the words you've successfully searched",
+    )
+
+    # Add sub-command s
+    parser_sw = sub_parsers.add_parser("s", help="search a word or phrase")
+
+    # Make sub-command s run default function of "search_words"
+    parser_sw.set_defaults(func=search_words)
+
+    # Add positional arguments with n args
+    parser_sw.add_argument(
+        "words",
+        nargs="+",
+        help="A word or phrase you want to search",
+    )
+    # Add optional arguments
+    parser_sw.add_argument(
+        "-v",
+        "--verbose",
+        action="store_true",
+        help="search a word or phrase in verbose mode",
+    )
+
+    # If it is for help
+    if len(sys.argv)==1:
+        parser.print_help()
+        sys.exit(0)
+
+    # If it is to search words
+    elif sys.argv[1] != "l" and len(sys.argv) > 1:
+        if "-v" in sys.argv or "--verbose" in sys.argv:
+            args = parser_sw.parse_args(["-v",  " ". join([i for i in sys.argv[1:] if not "-v" in i])])
+        elif "-h" in sys.argv or "--help" in sys.argv:
+            parser_sw.print_help()
+            sys.exit(0)
+        else:
+            args = parser_sw.parse_args([" ". join(sys.argv[1:])])
+
+    # If it is not to search words
+    else:
+        args = parser.parse_args()
+
     return args
 
 
+def get_query_word(args):
+    input_word = args.words[0]
+    query_word = args.words[0].replace(" ", "-").replace("/", "-")
+    return input_word, query_word
+
+
 def fetch(url, session):
     headers = {
         "User-Agent": "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/29.0.1547.62 Safari/537.36"
     }
     session.headers.update(headers)
-    attempt = 0
+    attempt = 0 
+
     while True:
-        if attempt == 3:
-            logger.info("Maximum amount of url fetch retries reached")
-            sys.exit(0)
         try:
-            r = session.get(url)
-            attempt += 1
-        except requests.exceptions.ConnectTimeout:
-            logger.error("Timed out during fetching %s. Retrying...", url)
-            time.sleep(0.1)
-        except requests.exceptions.ConnectionError:
-            logger.error("%s terminated connection. Retrying...", url)
-            time.sleep(0.1)
-        except Exception:
-            logger.exception("Error occured during fetching %s", url)
+            r = session.get(url, timeout=9.05)
+        except requests.exceptions.HTTPError as e:
+            attempt = call_on_error(e, url, attempt, OP[0])
+        except requests.exceptions.ConnectTimeout as e:
+            attempt = call_on_error(e, url, attempt, OP[0])
+        except requests.exceptions.ConnectionError as e:
+            attempt = call_on_error(e, url, attempt, OP[0])
+        except requests.exceptions.ReadTimeout as e:
+            attempt = call_on_error(e, url, attempt, OP[0])
+        except Exception as e:
+            attempt = call_on_error(e, url, attempt, OP[0])
         else:
             return r
 
 
+def call_on_error(error, url, attempt, op):
+    attempt += 1
+    logger.debug(
+        "%s HTML from %s %d times",
+        op,
+        url,
+        attempt,
+    )
+    if attempt == 3:
+        logger.debug("Maximum %s retries reached. Exit", op)
+        logger.error(str(error))
+        sys.exit()
+    return attempt
+
+
 def parse_spellcheck(args, session):
     word = " ".join(args.words).upper()
     query_word = "+".join(args.words)
@@ -74,35 +236,37 @@ def parse_spellcheck(args, session):
             console.print("[#b2b2b2]" + "  • " + suggestion)
 
 
-def parse_first_dict(response):
-    soup = BeautifulSoup(response.text, "lxml")
+def parse_first_dict():
+    global RESPONSE_TEXT
+    global RESPONSE_URL
+    if "isn’t in the Cambridge Dictionary yet. You can help!" in RESPONSE_TEXT or "404. Page not found" in RESPONSE_TEXT:
+        print("\n" + str(NoResultError()) + "\n")
+        sys.exit()
+    soup = BeautifulSoup(RESPONSE_TEXT, "lxml")
+    first_dict = soup.find("div", "pr dictionary")
+    if not first_dict:
+        print("\n" + str(ParsedNoneError()) + "\n")
+        sys.exit()
+    else:
+        return first_dict
 
-    attempt = 0
-    while True:
-        if attempt == 3:
-            logger.info("Something went wrong with beautifulsoup. Please try later")
-            sys.exit(0)
-        first_dict = soup.find("div", "pr dictionary")
-        attempt += 1
-        if not first_dict:
-            logger.error("Beautifulsoup found None after response for the word has been successfully fetched. Retrying...")
-            time.sleep(0.1)
-        else:
-            return first_dict
-
-
-def parse_dict_blocks(response):
-    first_dict = parse_first_dict(response)
-    if first_dict.find(
-        "div", ["pr entry-body__el", "entry-body__el clrd js-share-holder"]
-    ):
-        blocks = first_dict.find_all(
+def parse_dict_blocks():
+    first_dict = parse_first_dict()
+    try:
+        result = first_dict.find(
             "div", ["pr entry-body__el", "entry-body__el clrd js-share-holder"]
         )
+    except AttributeError:
+        print("\n" + str(NoResultError()) + "\n")
+        sys.exit()
     else:
-        blocks = first_dict.find_all("div", "pr idiom-block")
-    return blocks
-
+        if result:
+            blocks = first_dict.find_all(
+                "div", ["pr entry-body__el", "entry-body__el clrd js-share-holder"]
+            )
+        else:
+            blocks = first_dict.find_all("div", "pr idiom-block")
+        return blocks, first_dict
 
 # ----------parse dict head----------
 def parse_head_title(title_block):
@@ -110,7 +274,6 @@ def parse_head_title(title_block):
         word = title_block.find("div", "di-title").text.upper()
         return word
 
-
 def parse_head_type(head):
     if head.find("span", "anc-info-head danc-info-head"):
         w_type = (
@@ -475,45 +638,58 @@ def parse_dict_body(block):
         parse_phrasal_verb(block)
 
 
-def parse_dict_name(response):
-    first_dict = parse_first_dict(response)
+def parse_dict_name(first_dict):
     dict_info = replace_all(first_dict.small.text).replace("(", "").replace(")", "")
     dict_name = dict_info.split("©")[0]
     dict_name = dict_name.split("the")[-1]
-    console.print("[#125D98]" + "\n\n" + dict_name + "\n", justify="right")
+    console.print("[#125D98]" + dict_name + "\n", justify="right")
+
+def request_and_print(url, args):
+    with requests.Session() as session:
+        response = fetch(url, session)
+        if response.url == DICT_BASE_URL:
+            parse_spellcheck(args, session)
+            return False
+        else:
+            global RESPONSE_WORD
+            global RESPONSE_URL
+            global RESPONSE_TEXT
+            RESPONSE_URL, RESPONSE_WORD = parse_from_url(response.url)
+            RESPONSE_TEXT = response.text
+            logger.debug("Fetched html from %s successfully", RESPONSE_URL)
+            print_dict()
+            return True
+
+def print_dict():
+    blocks, first_dict = parse_dict_blocks()
+    for block in blocks:
+        parse_dict_head(block)
+        parse_dict_body(block)
+    parse_dict_name(first_dict)
 
 
 # ----------main----------
-
-
 def main():
     try:
+        con = sqlite3.connect(DB, detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES)
+        cur = con.cursor()
+
         args = parse_args()
-        query_word = "-".join(args.words)
-        url = DICT_BASE_URL + query_word
-        with requests.Session() as session:
-            response = fetch(url, session)
-            if response.url == DICT_BASE_URL:
-                parse_spellcheck(args, session)
-            elif response.url != url and response.url != DICT_BASE_URL:
-                response = fetch(response.url, session)
-                for block in parse_dict_blocks(response):
-                    parse_dict_head(block)
-                    parse_dict_body(block)
-                parse_dict_name(response)
-            else:
-                for block in parse_dict_blocks(response):
-                    parse_dict_head(block)
-                    parse_dict_body(block)
-                parse_dict_name(response)
+        args.func(args, con, cur)
+
+        cur.close()
+        con.close()
+
     except KeyboardInterrupt:
         print("\nStopped by user")
 
 
 if __name__ == "__main__":
     from log import logger
-    from utils import replace_all
+    from utils import replace_all, parse_from_url
     from console import console
+    from errors import ParsedNoneError, NoResultError
+    from cache import DB, create_table, insert_into_table, get_cache, get_response_words, get_random_words, delete_word, check_table
 
     t1 = time.time()
     main()
@@ -523,5 +699,7 @@ if __name__ == "__main__":
 
 else:
     from .log import logger
-    from .utils import replace_all
+    from .utils import replace_all, parse_from_url
     from .console import console
+    from .errors import ParsedNoneError, NoResultError
+    from .cache import DB, create_table, insert_into_table, get_cache, get_response_words, get_random_words, delete_word, check_table
